-- Handy shortcuts
local wv = BeneCast.WorkVars;
local BeneCastButton = BeneCast.Button;
local BeneCastPanel = BeneCast.Panel;
local BeneCastPanelDropDown = BeneCast.Panel.DropDown;
local hp = BeneCast.Helper;
local L = BeneCast.Localisation;
local const = BeneCast.Constants;

-- Called when AceAddon detects an ADDON_LOADED event for BeneCast
function BeneCast:OnInitialize()

	-- The variables should be loaded
	self.db = BeneCastDB;
			
	-- Set BC_class to the player's class
	_, wv.PlayerClass = UnitClass('player');
	
	wv.DefaultConfig = self:GetDefaultConfig();
			
	-- Reset the data if BeneCastDB is severely outdated
	local reason;
	if not self.db then
		reason = 'BeneCast.db = nil';
	elseif (not self.db.Profiles or hp._CountMembers(self.db.Profiles) == 0) then
		reason = 'No profiles found';
	elseif (not self.db.Setup or hp._CountMembers(self.db.Setup) == 0) then
		reason = 'No Setup found';
	elseif (not self.db.Setup.version or (self.db.Setup.version < const.DBVersion)) then
		reason = 'DB.Version empty or too low';
	end
	if reason then
		self:Print('Resetting settings: '..reason);
		BeneCastDB = hp._CopyTable(wv.DefaultConfig,true);
		self.db = BeneCastDB;
	end
	wv.Setup = self.db.Setup;
	wv.Profiles = self.db.Profiles;
	wv.CustomButtons = BeneCast.db.CustomButtons;
	wv.CustomEffects = BeneCast.db.CustomEffects;
	
	self:FillSpellInfo();
	
	-- Load up spell data (and other data) once AddOn has been loaded
	wv.LoadKnownSpells = true;
	wv.LoadCustomButtons = true;
	wv.LoadPreloadButtons = true;
	self:ResetData();
	
	-- Restore minimapbutton
	if wv.MinimapButton then
		if wv.MinimapButton:GetMinimapButtonPosition() == nil then
			wv.MinimapButton:SetMinimapButtonPosition(random(0,360));
		end
		wv.MinimapButton:UpdatePosition();
		if wv.Setup.hideminimapbutton then
			wv.MinimapButton:Hide();
		else
			wv.MinimapButton:Show();
		end
	end
	
	self:RegisterEvent('VARIABLES_LOADED');
	
end

-- Called right after OnInitialize and on subsequent manual enables
function BeneCast:OnEnable()

	-- Register for Events
	self:RegisterEvent('PLAYER_ENTERING_WORLD');
	self:RegisterEvent('PLAYER_LEAVING_WORLD');
	self:RegisterEvent('SPELLS_CHANGED');
	self:RegisterEvent('PARTY_MEMBERS_CHANGED');
	self:RegisterEvent('PLAYER_TARGET_CHANGED');
	self:RegisterEvent('PLAYER_FOCUS_CHANGED');
	self:RegisterEvent('UNIT_AURA');
	self:RegisterEvent('PLAYER_AURAS_CHANGED');
	self:RegisterEvent('UNIT_AURASTATE');
	self:RegisterEvent('UPDATE_SHAPESHIFT_FORMS');
	self:RegisterEvent('PLAYER_ALIVE');
	self:RegisterEvent('PLAYER_DEAD');
	self:RegisterEvent('PLAYER_UNGHOST');
	self:RegisterEvent('UNIT_PET');
	self:RegisterEvent('UNIT_TARGET');
	self:RegisterEvent('PLAYER_REGEN_DISABLED');
	self:RegisterEvent('PLAYER_REGEN_ENABLED');
		
	-- Restore disabled timers
	for i, timername in pairs(wv.DisabledTimers) do
		BeneCast.Timers[timername]:Enable();
	end
	hp._ClearTable(wv.DisabledTimers, false);

end

function BeneCast:OnDisable()

	for timername, timer in pairs(BeneCast.Timers) do
		if timer.handle then
			table.insert(wv.DisabledTimers, timername);
		end
		timer:Disable();
	end
	self:CancelAllTimers(); -- just to be safe
	-- Unregister for Events
	self:UnregisterEvent('PLAYER_ENTERING_WORLD');
	self:UnregisterEvent('PLAYER_LEAVING_WORLD');
	self:UnregisterEvent('SPELLS_CHANGED');
	self:UnregisterEvent('PARTY_MEMBERS_CHANGED');
	self:UnregisterEvent('PLAYER_TARGET_CHANGED');
	self:UnregisterEvent('PLAYER_FOCUS_CHANGED');
	self:UnregisterEvent('UNIT_AURA');
	self:UnregisterEvent('PLAYER_AURAS_CHANGED');
	self:UnregisterEvent('UNIT_AURASTATE');
	self:UnregisterEvent('UPDATE_SHAPESHIFT_FORMS');
	self:UnregisterEvent('PLAYER_ALIVE');
	self:UnregisterEvent('PLAYER_DEAD');
	self:UnregisterEvent('PLAYER_UNGHOST');
	self:UnregisterEvent('UNIT_PET');
	self:UnregisterEvent('UNIT_TARGET');
	self:UnregisterEvent('PLAYER_REGEN_DISABLED');
	self:UnregisterEvent('PLAYER_REGEN_ENABLED');

end

function BeneCast:GetProfile(pf, unit)
	if not wv.Profiles then
		-- Addon hasn't initialized yet
		return;
	end
	
	assert(wv.Profiles[pf]);
	
	if not wv.Profiles[pf].snapto then
		wv.Profiles[pf].snapto = hp.new();
	end
	
	if unit then
		return wv.Profiles[pf], wv.Profiles[pf].snapto[unit];
	else
		return wv.Profiles[pf];
	end
end

function BeneCast:GetCurrentProfileID()
	if wv.ConsiderInCombat then
		return 'combat';
	else
		return 'ooc';
	end
end

-- Function to resize the buttons of a profile
function BeneCast:ResizeButtonsForProfile(pf)

	if wv.Panels[pf] == nil then
		wv.Panels[pf] = hp.new();
	end
	
	-- Resize the buttons of every panel
	for unit, panel in pairs (wv.Panels[pf]) do
		panel:ResizeButtons();
	end
	
end

-- Handy function to clear out all ButtonInfo- and EventInfo-members of a certain type
function BeneCast:ClearButtonInfoOfOrigin(origin)

	local k, v, k2, v2;
	local ck;
	
	k, v = next(wv.ButtonInfo);
	while k do
		k2, v2 = next(wv.ButtonInfo, k);
		if v.origin and v.origin == origin then
			if v.type == 'cure' then
				for curetype, curetab in pairs(wv.CureInfo) do
					ck = hp_GetKeyForVal(curetab, k);
					if ck then
						curetab[ck] = nil;
					end
				end
			end
			hp._ClearTable(v, true);
			hp.del(v);
			wv.ButtonInfo[k] = nil;
		end
		k, v = k2, v2;
	end
	
	k, v = next(wv.EffectInfo);
	while k do
		k2, v2 = next(wv.EffectInfo, k);
		if v.origin and v.origin == origin then
			hp._ClearTable(v, true);
			hp.del(v);
			wv.EffectInfo[k] = nil;
		end
		k, v = k2, v2;
	end

end

-- Resets data tables
function BeneCast:ResetData()

	-- Rebuild spells
	local updateoptions;
	
	-- Clearing first, before adding buttons. It'll go faster when the tables are still empty.
	if wv.LoadKnownSpells then
		self:ClearButtonInfoOfOrigin('spell');
		updateoptions = true;
	end
	if wv.LoadCustomButtons then
		self:ClearButtonInfoOfOrigin('custom');
		updateoptions = true;
	end
	if wv.LoadPreloadButtons or updateoptions then
		self:ClearButtonInfoOfOrigin('preload');
		updateoptions = true;
	end
	
	-- Load the spell data from the spellbook
	if wv.LoadKnownSpells then
		self:ClearKnownSpells();
		self:LoadSpellData();
		-- Rebuild the button-definitions for spells
		self:AddSpellsToButtons();
	end
	
	-- TODO: add item-handling
	
	-- Rebuild preloadbuttons
	if wv.LoadPreloadButtons then
		for name, button in pairs(BeneCast.PreloadButtons) do
			if button.class == nil or (button.class ~= nil and hp._GetKeyForVal(button.class,wv.PlayerClass)) then
				wv.ButtonInfo[name] = hp._CopyTable(button, true);
				if button.type == 'cure' and (button.cures ~= nil and type(button.cures) == 'table') then
					for i, curetype in pairs(button.cures) do
						tinsert(wv.CureInfo[curetype], name);
					end
				end
			end
		end
		for name, effect in pairs(BeneCast.PreloadEffects) do
			if effect.class == nil or (effect.class ~= nil and hp._GetKeyForVal(effect.class,wv.PlayerClass)) then
				wv.EffectInfo[name] = hp._CopyTable(effect, true);
			end
		end
		wv.LoadPreloadButtons = false;
	end
	
	-- Rebuild custom buttons
	-- Doing this last allows a Custom Button to overwrite non-custom one
	-- Because of this, if one of the above is reloaded, the custom buttons also need to reload
	if wv.LoadCustomButtons or updateoptions then
		local v, oldorigin;
		-- Rebuild the button-definitions for custom buttons
		for name, button in pairs(wv.CustomButtons) do
			oldorigin = nil;
			v = wv.ButtonInfo[name];
			if v then
				oldorigin = v.origin;
				-- if overwiting we need to clean up CureInfo too
				if v.type == 'cure' then
					for curetype, curetab in pairs(wv.CureInfo) do
						ck = hp_GetKeyForVal(curetab, k);
						if ck then
							curetab[ck] = nil;
						end
					end
				end
				hp._ClearTable(v, true);
				hp.del(v);
				wv.ButtonInfo[name] = nil;
				v = nil;
			end
			wv.ButtonInfo[name] = hp._CopyTable(button, true);
			wv.ButtonInfo[name].publish = nil; -- Publish info is stored in the db, not here
			wv.ButtonInfo[name].oldorigin = oldorigin; -- So we can revert the customization
			button.oldorigin = oldorigin; -- Update the saved version too
			if button.type == 'cure' and (button.cures ~= nil and type(button.cures) == 'table') then
				for i, curetype in pairs(button.cures) do
					tinsert(wv.CureInfo[curetype], name);
				end
			end
		end
		for name, effect in pairs(wv.CustomEffects) do
			oldorigin = nil;
			v = wv.EffectInfo[name];
			if v then
				oldorigin = v.origin;
				hp._ClearTable(v, true);
				hp.del(v);
				wv.EffectInfo[name] = nil;
				v = nil;
			end
			wv.EffectInfo[name] = hp._CopyTable(effect, true);
			wv.EffectInfo[name].publish = nil; -- Publish info is stored in the db, not here
			wv.EffectInfo[name].oldorigin = oldorigin; -- So we can revert the customization
			effect.oldorigin = oldorigin; -- Update the saved version too
		end
		wv.LoadCustomButtons = false;
	end
	
	if updateoptions then
		-- Only rebuild optionstable if things have changed
		-- I'm feeding the old table to the garbage collector because I don't wanna risk destroying tables that might be linked by other code
		wv.Options = hp._CopyTable(wv.DefaultOptions, true);
		self:GenerateAllConfigs(wv.Options);
		LibStub('AceConfig-3.0'):RegisterOptionsTable('BeneCast', wv.Options, {'benecast', 'bc'});
	else
		self:Debug('BeneCast:ResetData(): Nothing updated');
	end
	
end

-- Function to make a shiny new, blank config
--  Doesn't mean that it will be used to actually reset the config but one does need it around
function BeneCast:GetDefaultConfig()

	-- Reset the saved table
	local newconfig = {
		Setup = {
			debug = nil,
			version = const.DBVersion,
			snapto = 'BeneCast',
			hideminimapbutton = false,
			unlockbuttons = false,
			-- Stores the position of the minimap button (needed for the minimap-code I swiped from Bongos2)
			minimappos = 0,
		},
		Profiles = {
			ooc = {
				name = L['Out of Combat'],
			},
			combat = {
				name = L['In Combat'],
			},
		},
		CustomButtons = {},
		CustomEffects = {},
	};
	
	self:ResetConfig(newconfig.Profiles.ooc, 'ooc');
	self:ResetConfig(newconfig.Profiles.combat, 'combat');
	
	return newconfig;
	
end

function BeneCast:ResetConfig(profile, defaultname)

	local profname = profile.name;
	hp._ClearTable(profile, true);
	if profname == nil then
		profname = defaultname;
	end 
	profile.name = profname;
	
	profile.snapto = hp.new(); -- this will store the panel's positions, k = unit, v = { x = ??, y = ??, ... }
	profile.config = hp.new(); -- this will store the spells configurations, k = groupname, v = { enabled = ?? }
	for i, v in pairs(const.Configs) do
		profile.config[v] = hp.new();
	end

	profile.showallbuttons = false;
	profile.flashasfade = false;
	profile.showtooltips = true;
	profile.showtooltipname = false;
	profile.cascadeplayer = true;
	profile.showpets = false;
	profile.checkrange = false; 
	profile.showtargettarget = true;
	profile.buttonsize = 36;
	profile.orientation = 'RIGHT';
	profile.unlockbuttons = false;

end

function BeneCast:Debug(msg)

	if wv.Setup and wv.Setup.debug then
		self:Print('Debug: ' .. msg);
	end

end

function BeneCast:AddButtonThatFlashes(button)

	if not mv.ButtonsThatFlash[pf] then
		mv.ButtonsThatFlash[pf] = hp.new();
	end
	if not mv.ButtonsThatFlash[pf][unit] then
		mv.ButtonsThatFlash[pf][unit] = hp.new();
	end
	table.insert(mv.ButtonsThatFlash[pf][unit],button);
	BeneCast.Timers.FlashButtons:Enable();

end

function BeneCast:UpdateAllPanels()

	for pfid, pfpanels in pairs(wv.Panels) do
		for unit, panel in pairs(pfpanels) do
			panel:Update();
		end
	end

end

function BeneCast:UpdatePanelsForUnit(unit)
	
	if unit then
		local panel;
		for pfid, pfpanels in pairs(wv.Panels) do
			panel = pfpanels[unit];
			if panel then
				panel:Update();
			end
		end
	end
	
end

function BeneCast:UpdatePanelsForProfile(pfid)
	
	if pfid then
		if wv.Panels[pfid] == nil then
			wv.Panels[pfid] = hp.new();
		end
	
		-- Update the buttons of every panel
		for unit, panel in pairs (wv.Panels[pfid]) do
			panel:Update(i);
		end
	end
	
end

function BeneCast:AddUnitToUpdate(unit, when)

	local utu = wv.UnitsToUpdate;
	if utu[unit] ~= nil then
		utu[unit] = when;
	else
		utu[unit] = when;
		utu.count = utu.count + 1;
	end
	if utu.count > 0 then
		BeneCast.Timers.UpdatePanels:Enable();
	end

end